# -*- coding: utf-8 -*-
"""Isotopologue correction for MS data sets.

Modules:
    isotope_correction.py: main function
    isotope_probabilities.py: subroutines and helper functions
    resolution_correction.py: subroutines for resolution depend correction
"""
from picor.isotope_correction import calc_isotopologue_correction
